@@ -30,7 +30,10 @@ use crate::raft::{
 };
 use crate::replication::{RaftEvent, ReplicaEvent, ReplicationStream};
 use crate::storage::HardState;
-use crate::{AppData, AppDataResponse, NodeId, RaftNetwork, RaftStorage};
+use crate::{
+    AppData, AppDataResponse, CatchUpCancellationPolicy, NodeId, RaftNetwork, RaftStorage,
+};
+use std::sync::atomic::{AtomicBool, Ordering};
 
 /// The core type implementing the Raft protocol.
 pub struct RaftCore<D: AppData, R: AppDataResponse, N: RaftNetwork<D>, S: RaftStorage<D, R>> {
@@ -198,17 +201,18 @@ impl<D: AppData, R: AppDataResponse, N: RaftNetwork<D>, S: RaftStorage<D, R>> Ra
         // Set initial state based on state recovered from disk.
         let is_only_configured_member =
             self.membership.members.len() == 1 && self.membership.contains(&self.id);
+
         // If this is the only configured member and there is live state, then this is
         // a single-node cluster. Become leader.
         if is_only_configured_member && self.last_log_index != u64::min_value() {
-            self.target_state = State::Leader;
+            self.set_target_state(State::Leader);
         }
         // Else if there are other members, that can only mean that state was recovered. Become follower.
         // Here we use a 30 second overhead on the initial next_election_timeout. This is because we need
         // to ensure that restarted nodes don't disrupt a stable cluster by timing out and driving up their
         // term before network communication is established.
         else if !is_only_configured_member {
-            self.target_state = State::Follower;
+            self.set_target_state(State::Follower);
             let inst = Instant::now()
                 + Duration::from_secs(30)
                 + Duration::from_millis(self.config.new_rand_election_timeout());
@@ -216,7 +220,7 @@ impl<D: AppData, R: AppDataResponse, N: RaftNetwork<D>, S: RaftStorage<D, R>> Ra
         }
         // Else, for any other condition, stay non-voter.
         else {
-            self.target_state = State::NonVoter;
+            self.set_target_state(State::NonVoter);
         }
 
         // This is central loop of the system. The Raft core assumes a few different roles based
@@ -273,8 +277,9 @@ impl<D: AppData, R: AppDataResponse, N: RaftNetwork<D>, S: RaftStorage<D, R>> Ra
     fn set_target_state(&mut self, target_state: State) {
         if target_state == State::Follower && !self.membership.contains(&self.id) {
             self.target_state = State::NonVoter;
+        } else {
+            self.target_state = target_state;
         }
-        self.target_state = target_state;
     }
 
     /// Get the next election timeout, generating a new value if not set.
@@ -594,6 +599,12 @@ struct LeaderState<'a, D: AppData, R: AppDataResponse, N: RaftNetwork<D>, S: Raf
     /// Once this receiver fires, the config change protocol can proceed and finish.
     pub(super) config_change_committed_cb:
         FuturesOrdered<oneshot::Receiver<Result<u64, RaftError>>>,
+    /// An optional receiver for when to cancel the catch up process of a
+    /// configuration change.
+    ///
+    /// If this receiver fires, then we should cancel the configuration change, and
+    /// go back to `Uniform` state.
+    pub(super) cancel_catch_up_cb: FuturesOrdered<oneshot::Receiver<bool>>,
 }
 
 impl<'a, D: AppData, R: AppDataResponse, N: RaftNetwork<D>, S: RaftStorage<D, R>>
@@ -614,6 +625,7 @@ impl<'a, D: AppData, R: AppDataResponse, N: RaftNetwork<D>, S: RaftStorage<D, R>
             awaiting_committed: Vec::new(),
             config_change_committed_cb: FuturesOrdered::new(),
             config_change_done_cb: None,
+            cancel_catch_up_cb: FuturesOrdered::new(),
         }
     }
 
@@ -706,6 +718,11 @@ impl<'a, D: AppData, R: AppDataResponse, N: RaftNetwork<D>, S: RaftStorage<D, R>
                     let _ = self.core.handle_replicate_to_sm_result(repl_sm_result);
                 }
                 Ok(_) = &mut self.core.rx_shutdown => self.core.set_target_state(State::Shutdown),
+                Some(Ok(should_cancel)) = self.cancel_catch_up_cb.next() => {
+                    if should_cancel {
+                        self.cancel_catch_up_state();
+                    }
+                }
             }
         }
     }
@@ -730,6 +747,36 @@ struct NonVoterReplicationState<D: AppData> {
     pub tx: Option<oneshot::Sender<Result<(), ChangeConfigError>>>,
 }
 
+/// A stateful instantiation of a `CatchUpCancellationPolicy`.
+///
+/// Used for cancellation-related bookkeeping.
+pub enum CatchUpCancellationState {
+    Timeout { already_caught_up: Arc<AtomicBool> },
+}
+
+impl CatchUpCancellationState {
+    fn new(cancel_tx: oneshot::Sender<bool>, policy: &CatchUpCancellationPolicy) -> Self {
+        match policy {
+            CatchUpCancellationPolicy::Timeout {
+                timeout_milliseconds,
+            } => {
+                let already_caught_up = Arc::new(AtomicBool::new(false));
+
+                let moved_timeout_milliseconds = *timeout_milliseconds;
+                let moved_already_caught_up = Arc::clone(&already_caught_up);
+                tokio::spawn(async move {
+                    tokio::time::sleep(Duration::from_millis(moved_timeout_milliseconds)).await;
+
+                    let should_cancel = !moved_already_caught_up.load(Ordering::Acquire);
+                    let _ = cancel_tx.send(should_cancel);
+                });
+
+                CatchUpCancellationState::Timeout { already_caught_up }
+            }
+        }
+    }
+}
+
 /// The nature of a configuration change.
 #[derive(Clone, Copy, PartialEq)]
 pub enum ConfigChangeOperation {
@@ -754,6 +801,7 @@ pub enum ConsensusState {
     CatchingUp {
         node: NodeId,
         tx: ChangeMembershipTx,
+        cancellation_state: CatchUpCancellationState,
     },
 
     /// A configuration change is in progress.
